--------------------------------------------------------------------------------
-- Copyright  2021								|
-- All Rights Reserved.								|
-- 										|
-- Permission to use, copy, modify, and distribute this software in any way	|
-- is not granted without expressed written agreement. This copyright applies	|
-- to the all code, data and artwork related to this project and all associated	|
-- projects published by the author.						|
--										|
-- More information on copyright is available from sources on-line.		|
-- Use an actual factual one!							|
--------------------------------------------------------------------------------

local AddonName, NS = ...
local L = NS.TEXTS

local ceil, floor = math.ceil, math.floor
local find, gsub = string.find, string.gsub
local tinsert, wipe = table.insert, table.wipe

local ActiveItems = 0
local ColumnCount = 3
local ItemsPerCol = 10
local ItemPrefix = AddonName.."ContextButton"
local NumPages = 2
local OffsetX = 20
local OffsetY = 12
local TimeToHide = 0.33

local ContextItems = {}
local FreeLabels = {}
local LabelList = {}
local OnCooldown = {}
local ColumnWidth, ItemHeight, LoadContextMenu, LoadLabels, MenuFrame, Options

local ContextTypes = {
	spell={ Drag=PickupSpell, CursorID=4, Info=GetSpellInfo, Cooldown=GetSpellCooldown, Tooltip=GameTooltip.SetSpellByID},
	item={ Drag=PickupItem, CursorID=2, Info=GetItemInfo, Cooldown=C_Container.GetItemCooldown, Tooltip=GameTooltip.SetItemByID},
	macro={ Drag=PickupMacro, CursorID=2, Info=GetMacroInfo },
-- Classic	mount={ Drag=_G.C_MountJournal.Pickup, CursorID=2, Info=C_MountJournal.GetMountInfo },
-- Classic	battlepet={ Drag=C_PetJournal.PickupPet, CursorID=2, Info=C_PetJournal.GetPetInfoByPetID, Summon=C_PetJournal.SummonPetByGUID },
}

local HearthstoneList = {
	[6948]="item", -- Hearthstone
	[64488]="item", -- "The Innkeeper's Daughter
	[28585]="item", -- "Ruby Slippers
	[54452]="item", -- "Ethereal Portal
	[93672]="item", -- "Dark Portal
	[142542]="item", -- "Tome of Town Portal
	[142543]="item", -- "Scroll of Town Portal
	[163694]="item", -- "Scroll of Luxurious Recall
	[37118]="item", -- Scroll of Recall
	[44314]="item", -- Scroll of Recall II (Level 35)
	[44315]="item", -- Scroll of Recall III (Level 65)
	[556]="spell", -- Astral Recall
	[209035]="item", -- flame-hearthstone
}

local function TakeOptions(optionstable) -- Used by FM Options plus drag-n-drop
	if Options then return end
	Options = optionstable
	Options.ColumnCount = ColumnCount
	Options.ColumnWidth = ColumnWidth
	Options.ContextTypes = ContextTypes
	Options.ItemHeight = ItemHeight
	Options.ItemsPerCol = ItemsPerCol
	Options.LoadContextMenu = LoadContextMenu
	Options.LoadLabels = LoadLabels
end
NS:AddOption("ContextFuncs", TakeOptions)

-------------------------------------------------------------------------------------------------------------------------------------------------------------
--						Local Functions
-------------------------------------------------------------------------------------------------------------------------------------------------------------
local function StripName(name, itemtype, id)
	if not name then
		NS.DebugPrint("OO can't StripName:", name, itemtype, id)
		return
	end
	local listtype = HearthstoneList[id]
	if listtype and listtype == itemtype  then
		name = GetBindLocation()
	else
		for usname, localename in pairs(L.ContextRemove) do -- the L.ContextRemove table might need expanding
			name = gsub(name, localename, "")
		end
	end
	return name
end

local function ToggleContextmenu(configmode, show)
	local Menu = NS:GetFrame("Context", "Parent")
	local settings = NS:GetSettings("Menu")
	if not settings.enabled then
		Menu.ConfigMode = nil
		Menu:Hide()
		return
	end
	if configmode then
		Menu.ConfigMode = show
		if show then
			Menu:Show()
		else
			Menu:Hide()
		end
		return
	end
	if settings.page1 then
		Menu:GetScript("OnMouseWheel")(Menu, 1)
	end
	Menu:Show()

--NS:GetFrame("Base", 4):Show() -- Fizz
end
NS:AddOption("ToggleContextmenu", ToggleContextmenu)

local function ResetCooldown(self)
	self.endcooldown = nil
	self.CDBackground:Hide()
	self.Cooldown:SetText("")
	self.Text:SetAlpha(1)
	self.Icon:SetAlpha(1)
	OnCooldown[self] = nil
end

local counter = 0
local function CooldownUpdate(self)
	local time = GetTime()
	for frame, timeleft in pairs(OnCooldown) do
		if MenuFrame.ConfigMode then
			if frame.endcooldown then
				ResetCooldown(frame)
			end
		else
			if time and frame.endcooldown and time > frame.endcooldown then
				ResetCooldown(frame)
			else
				frame.Cooldown:SetText(NS:TimeLeft(time, frame.endcooldown))
				frame.Text:SetAlpha(0.6)
				frame.Icon:SetAlpha(0.6)
				frame.CDBackground:Show()
			end
		end
	end
	if not next(OnCooldown) then
		NS:UnregisterForUpdates(MenuFrame, 5)
	end
end

local function CheckCooldown(self)
	self.endcooldown = nil
	local cooldownfromtable = ContextTypes[self:GetAttribute("type")].Cooldown
	if cooldownfromtable then
		local start, duration, enabled = cooldownfromtable(self.FMContextInfo)
		if (start and start > 0) and duration > 1.5 and enabled ~= 0 then -- < 1.5 mostly means on 1.46 GCD
			self.endcooldown = start+duration
			OnCooldown[self] = self.endcooldown
		end
	end
end

local function CheckForCooldowns()
	if next(OnCooldown) then
		CooldownUpdate()
	end
	for i=1, ActiveItems do
		CheckCooldown(ContextItems[i])
	end
	local register = next(OnCooldown)
	if register then
		NS:RegisterForUpdates(MenuFrame, 5, CooldownUpdate)
	else
		NS:UnregisterForUpdates(MenuFrame, 5)
	end
end

local function CreatePage(parent, page)
	local f = CreateFrame("Frame", "$parentPage"..page, parent, "SecureHandlerBaseTemplate")
	f:ClearAllPoints()
	f:SetPoint("TOPLEFT", 0, -5)
	f:SetPoint("BOTTOMRIGHT", 0, 3)
	f:Hide()
	f:SetFrameLevel(5)
	f.Empty = f:CreateFontString()
	f.Empty:SetFont(NS.Font, NS.FontSize+4, "")
	f.Empty:SetJustifyH("CENTER")
--	f.Empty:SetJustifyV("CENTER")
	f.Empty:SetJustifyV("MIDDLE")
	f.Empty:SetTextColor(0.85, 0.65, 0)--(1, 1, 0)
	f.Empty:SetPoint("CENTER")
	f.Empty:SetText(L.Empty)
	f.Empty:Hide()
	f:SetFrameRef("Parent", parent)
	f:SetAttribute("OffsetX", OffsetX)
	f:SetAttribute("OffsetY", OffsetY)
	f:SetAttribute("ColumnWidth", ColumnWidth)
--	f:SetFrameRef("Hiders", NS:GetFrame("Base", 4))
	f:Execute([[
		ItemTable = table.new()
		LastInCol = table.new()
		Close = [=[
				self:GetFrameRef("Parent"):Hide()
		]=]
	]])
--	if page == 1 then
--		f:Execute([[
--			ShowParent = [=[
--				self:GetFrameRef("Hiders"):Show()
--			]=]
--			HideParent = [=[
--				self:GetFrameRef("Hiders"):Hide()
--			]=]
--		]])
--	end
	return f
end

local function OnDragStart(self, ...)
	if not MenuFrame.ConfigMode then return end
	Options.TakeDrag(self)
end

local function OnReceiveDrag(self, ...) 
	if not MenuFrame.ConfigMode then return end
	Options.TakeDrop(self, ...)
end

local function ItemOnEnter(self, motion) 
	MenuFrame:GetScript("OnEnter")(MenuFrame, motion)
	local itemtype = self:GetAttribute("type")
	if not itemtype then return end
	local ttfunc = ContextTypes[itemtype].Tooltip
	if ttfunc then 
		if GetCVar("UberTooltips") == "1" then
			GameTooltip_SetDefaultAnchor(GameTooltip, self)
		else
			GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
		end
		GameTooltip:ClearLines()
		ttfunc(GameTooltip, self.FMContextInfo)
	end
end

local function ItemOnLeave(self, motion) 
	MenuFrame:GetScript("OnLeave")(MenuFrame, motion)
	GameTooltip:Hide()
end

local function ItemGet(id, attributes)
	local f = ContextItems[id] 
	if f then return f end
	f = CreateFrame("Button", ItemPrefix..id, UIParent, "SecureActionButtonTemplate, SecureHandlerBaseTemplate")
	ContextItems[id] = f
	f:SetSize(ColumnWidth, ItemHeight)
	f:SetHighlightFontObject("GameFontHighlightSmall")
	f:SetDisabledFontObject("GameFontDisableSmall")
	f:SetHighlightTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight")
	local hl = f:GetHighlightTexture()
	hl:SetVertexColor(1, 1, 0, 0.4)
	local layer = "OVERLAY"
	f.Icon = f:CreateTexture(nil, layer, nil, 1)
	f.Icon:SetSize(ItemHeight-2, ItemHeight-2)
	f.Icon:SetPoint("LEFT", 0, 0)

	f.CDBackground = f:CreateTexture(nil, layer, nil, 1)
	f.CDBackground:SetSize(5, 5)
	f.CDBackground:SetPoint("BOTTOMRIGHT", -2, 2)
	f.CDBackground:SetPoint("TOPLEFT", f, "TOP", f:GetWidth()/6, -f:GetHeight()/3)
	f.CDBackground:SetTexture(NS.TexturePath.."ContextCD")
	f.CDBackground:SetVertexColor(0.82, 0.51, 0)
	f.CDBackground:Hide()
	f.Cooldown = f:CreateFontString(nil, layer)
	f.Cooldown:SetFont(NS.Font, NS.FontSize-2.75, "")
	f.Cooldown:SetJustifyH("RIGHT")
--	f.Cooldown:SetJustifyV("CENTER")
	f.Cooldown:SetJustifyV("MIDDLE")
	f.Cooldown:SetText("00M")
	f.Cooldown:SetPoint("BOTTOMRIGHT", f, -3, 2)
	f.Stacks = f:CreateFontString()
	f.Stacks:SetDrawLayer(layer, 4)
	f.Stacks:SetFont(NS.Font, NS.FontSize-3, "OUTLINE")
	f.Stacks:SetJustifyH("RIGHT")
	f.Stacks:SetJustifyV("BOTTOM")
--	f.Stacks:SetPoint("TOPRIGHT", f, -1, 0)
--	f.Stacks:SetPoint("BOTTOMRIGHT", f, -1, 0)
--	f.Stacks:SetShadowOffset(1, -1)
	f.Stacks:SetPoint("BOTTOMRIGHT", f.Icon, 0, 2)
	f.Text = f:CreateFontString()--, layer -- If we have labed, the can go above text but below CDBackground and Cooldown(or maybe above, whichever works best)
	f.Text:SetDrawLayer("BACKGROUND")
	f.Text:SetFont(NS.Font, NS.FontSize-3, "")
	f.Text:SetJustifyH("LEFT")
--	f.Text:SetJustifyV("CENTER")
	f.Text:SetJustifyV("MIDDLE")
	f:SetFontString(f.Text) 
	f.Text:SetPoint("TOPLEFT", f.Icon, "TOPRIGHT", 1, -1)
	f.Text:SetPoint("BOTTOMLEFT", f.Icon, "BOTTOMRIGHT", 1, 1)
--	f.Text:SetPoint("RIGHT", f.Stacks, "LEFT", -1, 0)
	f.Text:SetPoint("RIGHT", f, -1, 0)
	f.Cooldown:SetText("")
	f:RegisterForDrag("LeftButton", "RightButton")
	f:SetScript("OnReceiveDrag", OnReceiveDrag) 
	f:SetScript("OnDragStart", OnDragStart) 
	f:SetScript("OnEnter", ItemOnEnter)
	f:SetScript("OnLeave", ItemOnLeave) 
	MenuFrame.Pages[1]:WrapScript(f, "OnClick", [[control:Run(Close)]])
	return f
end

local function ItemGetInfo(id, itemtype)
	local name, texture, stacks = "", ""
	if itemtype == "spell" then
		name = GetSpellInfo(id)
		texture = GetSpellTexture(name)
--		stacks = GetSpellCharges(id)
		stacks = GetSpellCount(id)
		if stacks and stacks < 1 then
			stacks = nil
		end
	elseif itemtype == "item" then
			name, _, _, _, _, _, _, _, _, texture = GetItemInfo(id)
	elseif itemtype == "macro" then
		name, texture = GetMacroInfo(id)
	elseif itemtype == "mount" then
		name, _, texture = C_MountJournal.GetMountInfoByID(id)
--[[
	elseif itemtype == "battlepet" then
		local customName
		_, customName, _, _, _, _, _, name, texture = C_PetJournal.GetPetInfoByPetID(id)
		if customName then
			name = customName
		end
]]--
	end
	return name, texture, stacks
end

local function ItemUpdate(self, itemsettings)
	local itemtype = itemsettings and itemsettings.oftype
	self.FMContextColumn = itemsettings and itemsettings.column
	self.FMContextId = itemsettings and itemsettings.id
	self.FMContextPage = itemsettings and itemsettings.page
	self.FMContextInfo = itemsettings and itemsettings.info
	self:SetAttribute("type", itemtype)
	if self.attr2 then
		self:SetAttribute(self.attr2, nil)
		self.attr2 = nil
	end
	if itemtype then
		local name, icon, stacks = ItemGetInfo(itemsettings.info, itemtype)
		self.attr2 = "*"..itemtype.."1"
		self:SetAttribute(self.attr2, name)
		self:SetText(StripName(name, itemtype, self.FMContextInfo))
		self.Icon:SetTexture(icon)
		self.Stacks:SetText(stacks)
--		self.Cooldown:SetTextColor(unpack(attributes and attributes.cooldowncolor or { 0.772, 0.556, 0.545, 1 }))
		self.Cooldown:SetTextColor(unpack(attributes and attributes.cooldowncolor or { 0, 0, 0, 1 }))
--		self.Cooldown:SetTextColor(unpack(attributes and attributes.cooldowncolor or { 1, 0.51, 0, 1 }))
		self.Text:SetTextColor(unpack(attributes and attributes.textcolor or { 1, 1, 1, 1 }))
	end
	self:SetAttribute("Column", itemsettings and itemsettings.column)
	self.endcooldown = nil
end

local function SortFunc(a, b)
	local x, y = a.col, b.col
	if x ~= y then
		return x < y
	else
		return a.id < b.id
	end
end

local function SortPage(table)
	local tbl = {}
	local Count = #table
	if Count < 1 then return tbl end
	for i=1, Count do
		tinsert(tbl, { col=table.column, id=i, entry=table[i] })
	end
	sort(tbl, SortFunc)
	return tbl
end

local function LabelAnchor(self, parent)
	self:SetParent(parent or UIParent)
	self:ClearAllPoints()
	if not parent then
		self:SetPoint("TOP", UIParent, "BOTTOM", 0, -50)
		return
	end
	self:SetPoint("LEFT", parent.Icon, "TOPRIGHT", 5, 2)
	self:SetPoint("RIGHT", parent, "TOPRIGHT", -5, 0)
	self:SetPoint("LEFT", parent.Icon, "TOPRIGHT", 5, 2)
	self:SetPoint("RIGHT", parent, "TOPRIGHT", -5, 0)
end

local function LabelColor(self, color)
	self.Texture:SetColorTexture(unpack(color or { 0.8, 0.6, 0, 1 }))
end

local function LabelColorText(self, color)
	self.Text:SetTextColor(unpack(color or { 0, 0, 0, 1 }))
end

local function AddLabel(parent, text, attributes)
	local f
	if next(FreeLabels) then
		f = FreeLabels[1]
		tremove(FreeLabels, 1)
	else		
		f = CreateFrame("Frame")
		tinsert(LabelList, f)
		f:SetSize(5, 5)
		f.SetAnchor = LabelAnchor
		f.SetColor = LabelColor
		f.SetColorText = LabelColorText
		f.Texture = f:CreateTexture()
		f.Texture:SetAllPoints()
		f.Texture:SetTexture("Interface\\BUTTONS\\WHITE8X8")
		f.Text = f:CreateFontString()
		f.Text:SetFont(NS.Font, NS.FontSize-4, "OUTLINE, MONOCHROME") --"OUTLINE, THICKOUTLINE, MONOCHROME"
		f.Text:SetJustifyH("CENTER")
--		f.Text:SetJustifyV("CENTER")
		f.Text:SetJustifyV("MIDDLE")
		f.Text:SetPoint("LEFT")
		f.Text:SetPoint("RIGHT")
		f.Text:SetText("III")
		f:SetHeight(f.Text:GetHeight())
	end
	f:SetAnchor(parent)
	f.Text:SetText(text)
	f:SetColor(attributes.color)
	f:SetColorText(attributes.textcolor)
	return f
end

-------------------------------------------------------------------------------------------------------------------------------------------------------------
--									Other Functions
-------------------------------------------------------------------------------------------------------------------------------------------------------------

local function LoadLabels()
	local settings = NS:GetSettings("Menu").labels
	if not settings then return end -- Temp. for testing. Reove when added to defaults
	local labels = settings.itemlabels
	if not labels then return end
	local attributes = settings.attributes
	wipe(FreeLabels)
	for i=1, #LabelList do
		tinsert(FreeLabels, LabelList[i])
		LabelList[i]:SetAnchor()
	end
	for k, v in pairs(ContextItems) do
		local label = v.FMContextPage.."-"..v.FMContextId
		label = labels[label] 
		if label then
			AddLabel(v, label.text, attributes)
		end
	end
end

LoadContextMenu = function()
	for pagecount=1, #MenuFrame.Pages do
		local menupage = MenuFrame.Pages[pagecount]
		menupage:Execute([[
			for i=1, #ItemTable do
				local item = ItemTable[i]
				item:SetParent(UIParent)
				item:ClearAllPoints()
				item:SetPoint("TOP", UIParent, "BOTTOM", 0, -50)
				item:Hide()
			end
			table.wipe(ItemTable)
			table.wipe(LastInCol)
		]])
	end
	local settings = NS:GetSettings("Menu")
	if not settings.enabled then return end
	local settingspages = settings.pages
	local itemcounter = 0
	for pagecount=1, #settingspages do
		MenuFrame.Pages[pagecount].Empty:Show()
		if #settingspages[pagecount] > 0 then
			local sorted = SortPage(settingspages[pagecount])
			for sortedcount=1, #sorted do
				itemcounter = itemcounter + 1
				local settingitem = sorted[sortedcount]
				settingitem = { id=settingitem.id, page=pagecount, info=settingitem.entry.info, column=settingitem.entry.column, oftype=settingitem.entry.oftype }
				local item = ItemGet(itemcounter, settings.attributes)
				ItemUpdate(item, settingitem)
				MenuFrame.Pages[pagecount].Empty:Hide()
				MenuFrame.Pages[pagecount]:SetFrameRef("NewItem", item)
				MenuFrame.Pages[pagecount]:Execute([[
					local OffsetX = self:GetAttribute("OffsetX")
					local OffsetY = self:GetAttribute("OffsetY")
					local ColumnWidth = self:GetAttribute("ColumnWidth")
					local item = self:GetFrameRef("NewItem")
					table.insert(ItemTable, item)
					item:SetParent(self)
					local column = item:GetAttribute("Column")
					local columnoffset = column-1
					item:ClearAllPoints()
					if LastInCol[column] then
						item:SetPoint("TOP", LastInCol[column], "BOTTOM")
					else
						item:SetPoint("TOPLEFT", self, (columnoffset*ColumnWidth)+OffsetX, -OffsetY)
					end
					item:Show()
					LastInCol[column] = item
				]])
			end
		end
	end
	ActiveItems = itemcounter
	LoadLabels()
	CheckForCooldowns()
end

local function CreateContextMenu(self)
	local f = CreateFrame("Frame", nil, UIParent) -- Context Menu must be parented to UIParent for secure children to SetPoint succesfully
	NS:AddFrame("Context", "Parent", f)
	MenuFrame = f
	f:SetFrameLevel(4)
	local top = NS:GetFrame("HotSpots", 6)
	f:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", top:GetLeft(), top:GetTop()) -- Setpoint is to UIParent get required offsets from GetLeft(), GetRight() etc.
	f:SetPoint("BOTTOMRIGHT", UIParent)
	f.Background = NS:GetFrame("Base", 4)
	f.OffsetX = OffsetX
	f.OffsetY = OffsetY
	ColumnWidth = (f:GetWidth()-(2*OffsetX))/ColumnCount
	ItemHeight = (f:GetHeight()-(2*OffsetY))/ItemsPerCol
	f.Texture = f:CreateTexture()
	f.Texture:SetAllPoints()
--	f.Texture:SetTexCoord(0.163, 1, 0, 1) -- old FM2
	f.Texture:SetTexCoord(0.16875, 1, 0, 1)
--	RegisterStateDriver(f, "visibility", "[target=target, exists, harm, nodead] hide")
	RegisterStateDriver(f, "visibility", "[combat] hide")
	
	f.Pages = {}
	for i=1, NumPages do
		f.Pages[i] = CreatePage(f, i)
	end
	f.CurrentPage = 2
	f:EnableMouseWheel(true)
	f:SetScript("OnMouseWheel", function(self, delta)
		delta = self.CurrentPage + delta
		local Change = (delta < 2 and 2) or (delta > 1 and 1) 
		if Change ~= self.CurrentPage then
			self.Pages[self.CurrentPage]:Hide()
			self.Pages[Change]:Show()
			self.CurrentPage = Change
			if MenuFrame.ConfigMode then 
				Options.SetPos(1, true)
			end
		end
	end)
	f:GetScript("OnMouseWheel")(f, 1)
	function f.CheckForCooldowns()
		C_Timer.After(0.5, CheckForCooldowns)
	end
	f:SetScript("OnShow", function(self)
--		if not self.Loaded then
--			self.Loaded = true
--			LoadContextMenu()
--			return
--		end
		CheckForCooldowns()
		LoadContextMenu() -- update stack information
		NS:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED", self, f.CheckForCooldowns)
	end)
--	MenuFrame.Pages[1]:WrapScript(f, "OnShow", [[control:Run(HideParent)]])
	f:SetScript("OnHide", function(self)
		NS:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED", self, f.CheckForCooldowns)
	end)
--	MenuFrame.Pages[1]:WrapScript(f, "OnHide", [[control:Run(ShowParent)]])
	f:SetScript("OnEnter", function(self, motion)
		if self.ConfigMode then return end
		self:SetScript("OnUpdate", nil) -- turn off the OnLeave close timer
	end)
	f:SetScript("OnLeave", function(self, motion)
		if self:IsMouseOver() or self.ConfigMode then return end  -- GetMouseFocus() instead of IsMouseOver maybe???
		self.TimeRemaining = TimeToHide -- Turn on the OnLeave close timer
		self:SetScript("OnUpdate", function(self, elapsed)
			self.TimeRemaining = self.TimeRemaining - elapsed
			if self.TimeRemaining < 0 then
				if InCombatLockdown() then
					self.TimeRemaining = 0.7 -- Longer time to check because ... Combat!
					return
				end
				self:SetScript("OnUpdate", nil)	
				self:Hide()
			end
		end)
	end)
	function f:SetColor(r, g, b, a) self.Texture:SetVertexColor(r, g, b, a) end
	f:Hide()
end

NS:RegisterForInit("PLAYER_LOGIN", "CreateContextMenu", 1, CreateContextMenu)
